@@ -32,37 +32,18 @@ int pulse_check_connection(snd_pulse_t * p)
 	pa_context_state_t state;
 
 	assert(p);
-	assert(p->context);
-	assert(p->mainloop);
+
+	if (!p->context || !p->mainloop)
+		return -EBADFD;
 
 	state = pa_context_get_state(p->context);
 
-	if (state != PA_CONTEXT_READY)
+	if (!PA_CONTEXT_IS_GOOD(state))
 		return -EIO;
 
 	return 0;
 }
 
-void pulse_stream_state_cb(pa_stream * s, void *userdata)
-{
-	snd_pulse_t *p = userdata;
-
-	assert(s);
-	assert(p);
-
-	pa_threaded_mainloop_signal(p->mainloop, 0);
-}
-
-void pulse_stream_success_cb(pa_stream * s, int success, void *userdata)
-{
-	snd_pulse_t *p = userdata;
-
-	assert(s);
-	assert(p);
-
-	pa_threaded_mainloop_signal(p->mainloop, 0);
-}
-
 void pulse_context_success_cb(pa_context * c, int success, void *userdata)
 {
 	snd_pulse_t *p = userdata;
@@ -77,8 +58,6 @@ int pulse_wait_operation(snd_pulse_t * p, pa_operation * o)
 {
 	assert(p);
 	assert(o);
-	assert(p->state == PULSE_STATE_READY);
-	assert(p->mainloop);
 
 	for (;;) {
 		int err;
@@ -96,44 +75,19 @@ int pulse_wait_operation(snd_pulse_t * p, pa_operation * o)
 	return 0;
 }
 
-int pulse_wait_stream_state(snd_pulse_t * p, pa_stream * stream,
-			    pa_stream_state_t target)
-{
-	pa_stream_state_t state;
-
-	assert(p);
-	assert(stream);
-	assert(p->state == PULSE_STATE_READY);
-	assert(p->mainloop);
-
-	for (;;) {
-		int err;
-
-		err = pulse_check_connection(p);
-		if (err < 0)
-			return err;
-
-		state = pa_stream_get_state(stream);
-
-		if (state == target)
-			break;
-
-		if (state == PA_STREAM_FAILED ||
-		    state == PA_STREAM_TERMINATED)
-			return -EIO;
-
-		pa_threaded_mainloop_wait(p->mainloop);
-	}
-
-	return 0;
-}
-
 static void context_state_cb(pa_context * c, void *userdata)
 {
+	pa_context_state_t state;
 	snd_pulse_t *p = userdata;
 	assert(c);
 
-	switch (pa_context_get_state(c)) {
+	state = pa_context_get_state(c);
+
+	/* When we get disconnected, tell the process */
+	if (!PA_CONTEXT_IS_GOOD(state))
+		pulse_poll_activate(p);
+
+	switch (state) {
 	case PA_CONTEXT_READY:
 	case PA_CONTEXT_TERMINATED:
 	case PA_CONTEXT_FAILED:
@@ -171,8 +125,6 @@ snd_pulse_t *pulse_new(void)
 	if (!p)
 		return NULL;
 
-	p->state = PULSE_STATE_INIT;
-
 	if (pipe(fd)) {
 		free(p);
 		return NULL;
@@ -197,7 +149,9 @@ snd_pulse_t *pulse_new(void)
 
 	p->context =
 	    pa_context_new(pa_threaded_mainloop_get_api(p->mainloop), buf);
-	assert(p->context);
+
+	if (!p->context)
+		goto fail;
 
 	pa_context_set_state_callback(p->context, context_state_cb, p);
 
@@ -207,36 +161,29 @@ snd_pulse_t *pulse_new(void)
 	return p;
 
 fail:
-	if (p->mainloop)
-		pa_threaded_mainloop_stop(p->mainloop);
-
-	if (p->context)
-		pa_context_unref(p->context);
-
-	if (p->mainloop)
-		pa_threaded_mainloop_free(p->mainloop);
-
-	if (p->main_fd >= 0)
-		close(p->main_fd);
-
-	if (p->thread_fd >= 0)
-		close(p->thread_fd);
-
-	free(p);
+	pulse_free(p);
 
 	return NULL;
 }
 
 void pulse_free(snd_pulse_t * p)
 {
-	pa_threaded_mainloop_stop(p->mainloop);
+	if (p->mainloop)
+		pa_threaded_mainloop_stop(p->mainloop);
 
-	pa_context_disconnect(p->context);
-	pa_context_unref(p->context);
-	pa_threaded_mainloop_free(p->mainloop);
+	if (p->context) {
+		pa_context_disconnect(p->context);
+		pa_context_unref(p->context);
+	}
 
-	close(p->thread_fd);
-	close(p->main_fd);
+	if (p->mainloop)
+		pa_threaded_mainloop_free(p->mainloop);
+
+	if (p->thread_fd >= 0)
+		close(p->thread_fd);
+
+	if (p->main_fd >= 0)
+		close(p->main_fd);
 
 	free(p);
 }
@@ -244,11 +191,16 @@ void pulse_free(snd_pulse_t * p)
 int pulse_connect(snd_pulse_t * p, const char *server)
 {
 	int err;
+	pa_context_state_t state;
 
 	assert(p);
-	assert(p->context);
-	assert(p->mainloop);
-	assert(p->state == PULSE_STATE_INIT);
+
+	if (!p->context || !p->mainloop)
+		return -EBADFD;
+
+	state = pa_context_get_state(p->context);
+	if (state != PA_CONTEXT_UNCONNECTED)
+		return -EBADFD;
 
 	pa_threaded_mainloop_lock(p->mainloop);
 
@@ -256,15 +208,20 @@ int pulse_connect(snd_pulse_t * p, const char *server)
 	if (err < 0)
 		goto error;
 
-	pa_threaded_mainloop_wait(p->mainloop);
+	for (;;) {
+		pa_context_state_t state = pa_context_get_state(p->context);
 
-	if (pa_context_get_state(p->context) != PA_CONTEXT_READY)
-		goto error;
+		if (!PA_CONTEXT_IS_GOOD(state))
+			goto error;
+
+		if (state == PA_CONTEXT_READY)
+			break;
+
+		pa_threaded_mainloop_wait(p->mainloop);
+	}
 
 	pa_threaded_mainloop_unlock(p->mainloop);
 
-	p->state = PULSE_STATE_READY;
-
 	return 0;
 
       error:
